# keyword.py -- Keyword translator
# Gareth Rees, Ravenbrook Limited, 2001-02-21.
# $Id: //info.ravenbrook.com/project/p4dti/version/1.0/code/replicator/keyword.py#2 $
#
# This file is copyright (c) 2001 Perforce Software, Inc.  All rights  
# reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import re
import replicator
import string

# P4 keywords can't contain whitespace, hashes, or double quotes.
# Fields in P4 jobspecs also can't contain semicolons or slashes.  We
# need a one-to-one translation between these keywords and other
# strings.  We use the following translation:
#
# string  P4
# ' '     _    (space to underscore)
# '\'     \\   (we use backslash to escape so it must escape itself)
# '_'     \_   (underscore to backslash underscore)
# ';'     \:   (semicolon to backslash colon)
# '#'     \=   (hash to backslash equals)
# '/'     /|   (slash to backslash bar)
# '"'     \'   (double quote to backslash apostrophe)
#  c      \xab (where c is some other whitespace character and ab is the
#               hex representation).
class keyword_translator(replicator.translator):
    # fixed translations
    specials = [(' ', '_'),
                ('_', '\\_'),
                ('\\', '\\\\'),
                (';', '\\:'),
                ('/', '\\|'),
                ('#', '\\='),
                ('"', "\\'"),
             ]
    # tables of the fixed translations
    dt_to_p4 = {}
    p4_to_dt = {}

    def __init__(self):
        for (dt,p4) in self.specials:
            self.dt_to_p4[dt] = p4
            self.p4_to_dt[p4] = dt

    # translate a matched single character to an escape sequence:
    def char_to_p4(self, match):
        if self.dt_to_p4.has_key(match.group(0)):
            return self.dt_to_p4[match.group(0)]
        else:
            return '\\x%02x' % ord(match.group(0))

    # translate a matched escape sequence to a single character:
    def p4_to_char(self, match):
        if self.p4_to_dt.has_key(match.group(0)):
            return self.p4_to_dt[match.group(0)]
        else:
            return chr(string.atoi(match.group(1)[2:], 0x10))

    # translate to a p4 keyword
    def translate_0_to_1(self, s,
                         dt0 = None, dt1 = None,
                         issue0 = None, issue1 = None):
        return re.sub('[\\s_;/#"\\\\]', self.char_to_p4, s)

    # translate from a p4_keyword
    def translate_1_to_0(self, s,
                         dt0 = None, dt1 = None,
                         issue0 = None, issue1 = None):
        return re.sub("_|\\\\([_\\\\:|=']|x[0-9a-f]{2})", self.p4_to_char, s)

    def test(self):
        tests = 0
        failures = 0
        # single character tests
        for i in range(256):
            s = chr(i)
            e = self.translate_0_to_1(s)
            d = self.translate_1_to_0(e)
            tests = tests + 1
            if d != s:
                print "character %d failure: %s -> %s -> %s" % (i, s, e, d)
                failures = failures  + 1

        if failures == 0:
            print "Translator passed all single-character tests."

        # multi-character tests
        for s in ['the quick brown fox jumps over the lazy dog',
                  '4_1_1_RELEASE',
                  '\\TeX',
                  '(a;b;c)',
                  '"quote"',
                  ' la recherche du temps perdu',
                  "sop's fables",
                  "Ernesto Cesro, Pl Erds, ystein Ore, Muhammad ibn Ms al-Khwrizm",
                  'Windows 3.1',
                  'HP-UX',
                  'OSF/1',
                  '###urgent###',
                  ]:
            e = self.translate_0_to_1(s)
            d = self.translate_1_to_0(e)
            tests = tests + 1
            if d == s:
                print "success: ", s, "->", e
            else:
                print "failed: ", s, "->", e, "->", d
                failures = failures + 1

        print "Translator failed %d tests out of %d." % (failures, tests)

# B. Document History
# 
# 2001-02-21  GDR  Created.
#
# 2001-03-02  RB    Transferred copyright to Perforce under their license.
